/* BBAltTray - a BB4W plugin to access the system tray by bro@ms
   Copyright (C) 2004 kana <nicht AT s8 DOT xrea DOT con>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <BBApi-C.h>

#include "cpp_adaptor.h"


#define NUMBER_OF(array) (sizeof(array) / sizeof((array)[0]))

typedef struct {
	int top;
	int right;
	int bottom;
	int left;
} Padding;




/* Constants */

static const char BBTK_NAME[] = "BBAltTray";
static const char BBTK_VERSION[] = "0.0.2";
static const char BBTK_AUTHOR[] = "kana";
static const char BBTK_RELEASE[] = "2004-12-20/2005-04-06";
static const char BBTK_LINK[] = "http://nicht.s8.xrea.com/2004/12/BBAltTray";
static const char BBTK_EMAIL[] = "nicht AT s8 DOT xrea DOT con";

static const char BBTK_CLASS_NAME[] = "BBAltTray";
static const char BBTK_RC_NAME[] = "BBAltTray.rc";
static const UINT BBTK_MESSAGES[] = {BB_BROADCAST, BB_RECONFIGURE, 0};




/* Global variables */

	/**
	 * These are set in beginPlugin().
	 * And these will never be changed until endPlugin().
	 */
static HWND G_PluginWnd = NULL;
static HMODULE G_PluginModule = NULL;

	/* These are set in read_settings(). */
static int G_DefaultScreenWidth = 640;
static int G_DefaultScreenHeight = 480;

static Padding G_BasePadding = {0, 0, 0, 0};
static int G_BaseStyle = SN_MENUFRAME;
static BOOL G_BaseBorder = FALSE;
static Padding G_TipPadding = {2, 13, 2, 4};
static int G_TipStyle = SN_MENUHILITE;
static BOOL G_TipBorder = FALSE;
static int G_IconTipPadding = 4;
static int G_IconSize = 16;
static BOOL G_DoConvertNewLine = TRUE;
static char G_ConvertedNewLine[16] = " / ";

	/**
	 * The current state of BBAltTray.
	 *
	 *                 Next | Prev    Next | Prev
	 *                    ---->          ----.
	 * ----> (not selecting) (selecting)     |
	 * start              <----          <---'
	 *                    Cancel
	 *
	 *                    <----
	 *                     Send
	 */
static enum {
	STATE_NOT_SELECTING,
	STATE_SELECTING
} G_State = STATE_NOT_SELECTING;

	/* The index of the current selected icon (0 ... G_TraySize-1). */
static int G_SelectedIcon;

	/* The number of icons in the system tray (0 <= G_TraySize). */
static int G_TraySize;








/* Misc. functions */

static void
notice(const char* format, ...)
{
	char buf[80 * 25];
	va_list va;

	va_start(va, format);
	vsnprintf(buf, NUMBER_OF(buf), format, va);
	buf[NUMBER_OF(buf) - 1] = '\0';
	va_end(va);

	MessageBox( G_PluginWnd, buf, BBTK_NAME,
	            MB_OK | MB_SETFOREGROUND | MB_TOPMOST );
}




static BOOL
get_module_dir(char* buf, int size)
{
	char* bs;

	if (!GetModuleFileName(G_PluginModule, buf, size))
		return FALSE;
	buf[size - 1] = '\0';

	bs = strrchr(buf, '\\');
	if (bs != NULL)
		bs[1] = '\0';

	return TRUE;
}


static BOOL
get_rc_path(char* buf, int size, const char* rc_name)
{
	char dir[MAX_PATH];

	if (!get_module_dir(dir, NUMBER_OF(dir)))
		return FALSE;

	if (size < strlen(dir) + strlen(rc_name) + 1)
		return FALSE;

	strcpy(buf, dir);
	strcat(buf, rc_name);

	return TRUE;
}




static const char*
convert_new_line(char* _dest, const char* src, int size)
{
	char* last = _dest + (size - 1);
	char* dest = _dest;

	while (*src != '\0' && dest < last) {
		switch (*src) {
		default:
			*dest = *src;
			dest++;
			src++;
			break;

		case '\n':
			if ( (last - dest) / sizeof(*dest) - 1
			     <  strlen(G_ConvertedNewLine) )
			{
				goto END;
			}

			strcpy(dest, G_ConvertedNewLine);
			dest += strlen(G_ConvertedNewLine);
			src++;
			break;

		case '\r':
			src++;
			break;
		}
	}
END:
	*dest = '\0';
	return _dest;
}




static int
get_screen_width(void)
{
	int w;

	w = GetSystemMetrics(SM_CXSCREEN);
	if (!(0 < w))
		w = G_DefaultScreenWidth;

	return w;
}


static int
get_screen_height(void)
{
	int h;

	h = GetSystemMetrics(SM_CYSCREEN);
	if (!(0 < h))
		h = G_DefaultScreenHeight;

	return h;
}




static Padding
read_padding(LPCSTR rc, LPCSTR key, LPCSTR default_value)
{
	LPCSTR s;
	Padding p;

	s = ReadString(rc, key, default_value);

	switch (sscanf( s, "%d %d %d %d",
	                &(p.top), &(p.right), &(p.bottom), &(p.left) ))
	{
	default:
		p.top = 0;
	case 1:
		p.right = p.top;
	case 2:
		p.bottom = p.top;
	case 3:
		p.left = p.right;
	case 4:
		break;
	}

	return p;
}


static const struct {const char* name;  int number;} STYLE_TABLE[] = {
	{"toolbar", SN_TOOLBAR},
	{"toolbar.label", SN_TOOLBARLABEL},
	{"toolbar.windowlabel", SN_TOOLBARWINDOWLABEL},
	{"toolbar.clock", SN_TOOLBARCLOCK},
	{"menu.title", SN_MENUTITLE},
	{"menu.frame", SN_MENUFRAME},
	{"menu.hilite", SN_MENUHILITE},
};

static int
read_style(LPCSTR rc, LPCSTR key, LPCSTR default_value)
{
	LPCSTR s;
	int i;

	s = ReadString(rc, key, default_value);
	for (i = 0; i < NUMBER_OF(STYLE_TABLE); i++) {
		if (!stricmp(s, STYLE_TABLE[i].name))
			return STYLE_TABLE[i].number;
	}

	return SN_TOOLBAR;
}


static BOOL
read_bool(LPCSTR rc, LPCSTR key, LPCSTR default_value)
{
	return !stricmp("true", ReadString(rc, key, default_value));
}


static void
read_string(LPCSTR rc, LPCSTR key, LPCSTR default_value, LPSTR buf, int size)
{
	LPCSTR s;
	int l;
	int m;

	s = ReadString(rc, key, default_value);
	l = strlen(s);
	if ((s[0] == '"') && (s[l-1] == '"')) {
		s++;
		l--;

		l--;
	}

	m = ((l+1 < size) ? l+1: size);
	strncpy(buf, s, m);
	buf[m-1] = '\0';
}








/* Main functions */

#define RC_KEY(s) "bbalttray." s ":"
#define RC_KEY_SCREEN_WIDTH RC_KEY("screen_width")
#define RC_KEY_SCREEN_HEIGHT RC_KEY("screen_height")
#define RC_KEY_BASE_PADDING RC_KEY("base_padding")
#define RC_KEY_BASE_STYLE RC_KEY("base_style")
#define RC_KEY_BASE_BORDER RC_KEY("base_border")
#define RC_KEY_TIP_PADDING RC_KEY("tip_padding")
#define RC_KEY_TIP_STYLE RC_KEY("tip_style")
#define RC_KEY_TIP_BORDER RC_KEY("tip_border")
#define RC_KEY_ICON_TIP_PADDING RC_KEY("icon_tip_padding")
#define RC_KEY_ICON_SIZE RC_KEY("icon_size")
#define RC_KEY_DO_CONVERT_NEW_LINE RC_KEY("do_convert_new_line")
#define RC_KEY_CONVERTED_NEW_LINE RC_KEY("converted_new_line")


static void
read_settings(void)
{
	char rc[MAX_PATH];

	if (!get_rc_path(rc, NUMBER_OF(rc), BBTK_RC_NAME)) {
		notice("Failed to get the path to %s", BBTK_RC_NAME);
		return;
	}

	G_DefaultScreenWidth = ReadInt(rc, RC_KEY_SCREEN_WIDTH, 640);
	G_DefaultScreenHeight = ReadInt(rc, RC_KEY_SCREEN_HEIGHT, 480);
	G_BasePadding = read_padding(rc, RC_KEY_BASE_PADDING, "0 0 0 0");
	G_BaseStyle = read_style(rc, RC_KEY_BASE_STYLE, "menu.frame");
	G_BaseBorder = read_bool(rc, RC_KEY_BASE_BORDER, "false");
	G_TipPadding = read_padding(rc, RC_KEY_TIP_PADDING, "2 13 2 4");
	G_TipStyle = read_style(rc, RC_KEY_TIP_STYLE, "menu.hilite");
	G_TipBorder = read_bool(rc, RC_KEY_TIP_BORDER, "false");
	G_IconTipPadding = ReadInt(rc, RC_KEY_ICON_TIP_PADDING, 4);
	G_IconSize = ReadInt(rc, RC_KEY_ICON_SIZE, 16);
	if (G_IconSize < 0)
		G_IconSize = 0;
	G_DoConvertNewLine = read_bool(rc, RC_KEY_DO_CONVERT_NEW_LINE, "true");
	read_string( rc, RC_KEY_CONVERTED_NEW_LINE, " / ",
	             G_ConvertedNewLine, NUMBER_OF(G_ConvertedNewLine) );
}


/*
static void
write_settings(void)
{
	char rc[MAX_PATH];

	if (!get_rc_path(rc, NUMBER_OF(rc), BBTK_RC_NAME)) {
		notice("Failed to get the path to %s", BBTK_RC_NAME);
		return;
	}

	...
}
*/




/**
 *  ,----------- G_BasePadding (left)
 *  | ,--------- G_TipPadding (left)
 *  | |   ,----- G_IconSize
 *  | |   |  ,-- G_IconTipPadding
 *  | |   |  |               ,---- G_TipPadding (right)
 *  | |   |  |               | ,-- G_BasePadding (right)
 *  | |   |  |               | |
 *  v v   v  v               v v
 * ,----------------------------. ___ G_BasePadding (top)
 * | ,------------------------. | ___ G_TipPadding (top)
 * | | ,--------------------. | |
 * | | |    | |             | | |
 * | | |icon| | tip...      | | |
 * | | `--------------------' | | ___ G_TipPadding (bottom)
 * | |------------------------| | ___ G_TipPadding (top)
 * | | ,--------------------. | |
 * | | |    | |             | | |
 * : : :    : :             : : :
 */

static void
update_size_and_position(void)
{
	HDC hdc;
	RECT rect;
	int max_width;
	int max_height;
	int screen_width;
	int screen_height;
	int width;
	int height;
	int i;
	HGDIOBJ old_font;
	void* base_style = GetSettingPtr(G_BaseStyle);

	hdc = GetDC(G_PluginWnd);
	if (hdc == NULL)
		return;
	old_font = SelectObject(hdc, Cpp_CreateStyleFont(base_style));

	ZeroMemory(&rect, sizeof(rect));
	max_height = G_IconSize;
	max_width = 0;
	for (i = 0; i < G_TraySize; i++) {
		systemTray* icon;
		char buf[NUMBER_OF(icon->szTip)];
		const char* tip;

		icon = GetTrayIcon(i);
		if (icon == NULL)
			break;

		if (G_DoConvertNewLine)
			tip = convert_new_line(buf,icon->szTip,NUMBER_OF(buf));
		else
			tip = icon->szTip;
		DrawText( hdc, tip, -1, &rect,
			  DT_CALCRECT | DT_NOPREFIX | DT_SINGLELINE );
		if (max_height < rect.bottom)
			max_height = rect.bottom;
		if (max_width < rect.right)
			max_width = rect.right;
	}

	DeleteObject(SelectObject(hdc, old_font));
	ReleaseDC(G_PluginWnd, hdc);

	width = G_BasePadding.left
	          + G_IconSize + G_IconTipPadding
	          + (G_TipPadding.left + max_width + G_TipPadding.right)
	          + G_BasePadding.right;
	height = G_BasePadding.top
	           + (G_TipPadding.top + max_height + G_TipPadding.bottom)
	             * G_TraySize
	           + G_BasePadding.bottom;

	screen_width = get_screen_width();
	screen_height = get_screen_height();

	SetWindowPos(
	  G_PluginWnd,
	  NULL,
	  screen_width / 2 - width / 2,
	  screen_height / 2 - height / 2,
	  width,
	  height,
	  SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_NOZORDER
	);
}




static int
_GetTraySize(void)
{
	int i;
	int size;

	size = GetTraySize();
	if (!(0 <= size))
		size = 0;

	for (i = 0; i < size; i++) {
		systemTray* icon;

		icon = GetTrayIcon(i);
		if ((icon == NULL) || !IsWindow(icon->hWnd)) {
			SendMessage(GetBBWnd(), BB_CLEANTRAY, 0, 0);
			size = GetTraySize();
			if (!(0 <= size))
				size = 0;
			break;
		}
	}

	return size;
}

static void
action_non_selecting_to_selecting(void)
{
	G_State = STATE_SELECTING;
	G_SelectedIcon = 0;
	G_TraySize = _GetTraySize();

	update_size_and_position();
	ShowWindow(G_PluginWnd, SW_SHOW);
	SetForegroundWindow(G_PluginWnd);
}

static void
action_selecting_to_non_selecting(void)
{
	G_State = STATE_NOT_SELECTING;

	ShowWindow(G_PluginWnd, SW_HIDE);
}




#define POST(icon, msg)						\
	PostMessage( (icon)->hWnd, (icon)->uCallbackMessage,	\
		     (WPARAM)((icon)->uID), (LPARAM)(msg) )
static void
send_click_message(const char* name)
{
	systemTray* icon;

	icon = GetTrayIcon(G_SelectedIcon);
	if (icon == NULL)
		return;

	SetForegroundWindow(icon->hWnd);

	if (strcmp(name, "LeftClick") == 0) {
		POST(icon, WM_LBUTTONDOWN);
		POST(icon, WM_LBUTTONUP);
	} else if (strcmp(name, "RightClick") == 0) {
		POST(icon, WM_RBUTTONDOWN);
		POST(icon, WM_RBUTTONUP);
	} else if (strcmp(name, "MiddleClick") == 0) {
		POST(icon, WM_MBUTTONDOWN);
		POST(icon, WM_MBUTTONUP);
	} else if (strcmp(name, "LeftDoubleClick") == 0) {
		POST(icon, WM_LBUTTONDOWN);
		POST(icon, WM_LBUTTONUP);
		POST(icon, WM_LBUTTONDBLCLK);
		POST(icon, WM_LBUTTONUP);
	} else if (strcmp(name, "RightDoubleClick") == 0) {
		POST(icon, WM_RBUTTONDOWN);
		POST(icon, WM_RBUTTONUP);
		POST(icon, WM_RBUTTONDBLCLK);
		POST(icon, WM_RBUTTONUP);
	} else if (strcmp(name, "MiddleDoubleClick") == 0) {
		POST(icon, WM_MBUTTONDOWN);
		POST(icon, WM_MBUTTONUP);
		POST(icon, WM_MBUTTONDBLCLK);
		POST(icon, WM_MBUTTONUP);
	} else {
		notice("Unknown message: `%s'", name);
	}
}
#undef POST




static void
interpret_broam(const char* broam)
{
	if (strncmp(broam, "@BBAltTray.", 11) != 0)
		return;
	broam += 11;

	if (strcmp(broam, "Next") == 0) {
		if (G_State == STATE_NOT_SELECTING) {
			action_non_selecting_to_selecting();
			return;
		}

		G_SelectedIcon++;
		if (G_TraySize <= G_SelectedIcon) {
			if (G_TraySize != 0)
				G_SelectedIcon -= G_TraySize;
			else
				G_SelectedIcon = 0;
		}
		InvalidateRect(G_PluginWnd, NULL, TRUE);
	} else if (strcmp(broam, "Prev") == 0) {
		if (G_State == STATE_NOT_SELECTING) {
			action_non_selecting_to_selecting();
			G_SelectedIcon = G_TraySize;
		}

		G_SelectedIcon--;
		if (G_SelectedIcon < 0) {
			if (G_TraySize != 0)
				G_SelectedIcon += G_TraySize;
			else
				G_SelectedIcon = 0;
		}
		InvalidateRect(G_PluginWnd, NULL, TRUE);
	} else if (strcmp(broam, "Cancel") == 0) {
		if (G_State == STATE_SELECTING)
			action_selecting_to_non_selecting();
		else
			; /* ignore */
	} else if ((strncmp(broam, "Send", 4) == 0) && isspace(broam[5-1])) {
		if (G_State == STATE_SELECTING) {
			broam += 5;
			while (isspace(*broam))
				broam++;
			send_click_message(broam);
		} else
			; /* ignore */

		action_selecting_to_non_selecting();
	} else {
		notice("Unknown bro@m: `%s'", broam - 11);
	}
}




static LRESULT CALLBACK
BBTK_WndProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg) {
	default:
		break;


	case BB_BROADCAST:
		interpret_broam((const char*)lp);
		break;

	case BB_RECONFIGURE:
		read_settings();
		break;


	case WM_CREATE:
		SendMessage( GetBBWnd(), BB_REGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBTK_MESSAGES );
		break;

	case WM_DESTROY:
		SendMessage( GetBBWnd(), BB_UNREGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBTK_MESSAGES );
		break;

	case WM_PAINT: {
		PAINTSTRUCT ps;
		HDC hdc;
		RECT rect;
		int i;
		int height;
		void* base_style = GetSettingPtr(G_BaseStyle);
		void* task_style = GetSettingPtr(G_TipStyle);
		COLORREF normal_text = Cpp_StyleItem_TextColor(base_style);
		COLORREF hilite_text = Cpp_StyleItem_TextColor(task_style);
		HGDIOBJ old_font;

		hdc = BeginPaint(hwnd, &ps);

		GetClientRect(hwnd, &rect);
		Cpp_MakeStyleGradient(hdc, &rect, base_style, G_BaseBorder);

		height = rect.bottom - rect.top;
		height -= (G_BasePadding.top + G_BasePadding.bottom);
		height /= G_TraySize;
		height -= G_TipPadding.top + G_TipPadding.bottom;

		rect.left += G_BasePadding.left + G_TipPadding.left;
		rect.top += G_BasePadding.top;
		rect.right -= G_BasePadding.right + G_TipPadding.right;
		/* rect.bottom */

		old_font = SelectObject(hdc, Cpp_CreateStyleFont(base_style));
		SetTextColor(hdc, normal_text);
		SetBkMode(hdc, TRANSPARENT);
		for (i = 0; i < G_TraySize; i++) {
			systemTray* icon;
			char buf[NUMBER_OF(icon->szTip)];
			const char* tip;

			icon = GetTrayIcon(i);
			if (icon == NULL)
				break;

			if (i == G_SelectedIcon) {
				RECT hilite_rect = rect;

				hilite_rect.left -= G_TipPadding.left;
				hilite_rect.right += G_TipPadding.right;
				hilite_rect.bottom = G_TipPadding.top
				                       + hilite_rect.top
				                       + height
				                       + G_TipPadding.bottom;
				Cpp_MakeStyleGradient( hdc, &hilite_rect,
				                       task_style,
				                       G_TipBorder );
				SetTextColor(hdc, hilite_text);
			}

			rect.top += G_TipPadding.top;

			if (0 < G_IconSize) {
				DrawIconEx( hdc, rect.left, rect.top,
				            icon->hIcon,
				            G_IconSize, G_IconSize,
				            0, NULL, DI_NORMAL );
				rect.left += G_IconSize + G_IconTipPadding;
			}

			rect.bottom = rect.top + height;
			if (G_DoConvertNewLine)
			  tip=convert_new_line(buf,icon->szTip,NUMBER_OF(buf));
			else
			  tip = icon->szTip;
			DrawText( hdc, tip, -1, &rect,
			          DT_NOPREFIX | DT_SINGLELINE | DT_VCENTER );

			if (0 < G_IconSize) {
				rect.left -= G_IconSize + G_IconTipPadding;
			}

			rect.top += height + G_TipPadding.bottom;

			if (i == G_SelectedIcon) {
				SetTextColor(hdc, normal_text);
			}
		}
		DeleteObject(SelectObject(hdc, old_font));

		EndPaint(hwnd, &ps);
		} break;
	}

	return DefWindowProc(hwnd, msg, wp, lp);
}








/* Public functions */

DLL_EXPORT int
beginPlugin(HINSTANCE hplugininstance)
{
	WNDCLASS wc;

	G_PluginModule = hplugininstance;

	read_settings();

	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = BBTK_WndProc;
	wc.hInstance = hplugininstance;
	wc.lpszClassName = BBTK_CLASS_NAME;
	if (!RegisterClass(&wc))
		goto E_REGISTER_CLASS;

	G_PluginWnd = CreateWindowEx( WS_EX_TOOLWINDOW | WS_EX_TOPMOST,
	                              BBTK_CLASS_NAME,
	                              NULL,
	                              WS_POPUP,
	                              0, 0, 0, 0,
	                              NULL,
	                              NULL,
	                              hplugininstance,
	                              NULL );
	if (G_PluginWnd == NULL)
		goto E_CREATE_WINDOW;

	return 0;


E_CREATE_WINDOW:
	UnregisterClass(BBTK_CLASS_NAME, hplugininstance);
E_REGISTER_CLASS:
	return !0;
}




DLL_EXPORT void
endPlugin(HINSTANCE hplugininstance)
{
	/* write_settings(); */

	DestroyWindow(G_PluginWnd);
	UnregisterClass(BBTK_CLASS_NAME, hplugininstance);
}




DLL_EXPORT LPCSTR
pluginInfo(int field)
{
	switch (field) {
	default: /* FALLTHRU */
	case PLUGIN_NAME: return BBTK_NAME;
	case PLUGIN_VERSION: return BBTK_VERSION;
	case PLUGIN_AUTHOR: return BBTK_AUTHOR;
	case PLUGIN_RELEASE: return BBTK_RELEASE;
	case PLUGIN_LINK: return BBTK_LINK;
	case PLUGIN_EMAIL: return BBTK_EMAIL;
	}
}




/* __END__ */
